/*
 * TismTool © Tis Veugen, 2010-2012
 * Simulation, Verification
 *
 * TsmSupervisor.java
 * version 13
 */
package TsmRTE.TsmSupervisor;

import java.util.*;
import TsmRTE.TsmComponent.*;
import TsmRTE.TsmGen.*;

enum TsmScenItemKind
{
  RELEASE_CB_MSG,
  RELEASE_CB_RV2,
  RELEASE_MSG,
  RELEASE_RV2,
  NATIVE,
  CHOICE,
  QO;
}

class TsmScenItem
{
  public TsmScenItemKind  Kind;
  public int              Index;
  public TsmClass         Receiver;
  public TsmThreadData    ThreadData;

  @Override
  public String toString()
  {
    switch ( Kind )
    {
    case RELEASE_CB_MSG :
      return "CB_MSG";
    case RELEASE_CB_RV2 :
      return "CB_RV2";
    case RELEASE_MSG :
      return "MSG " + ThreadData.ThreadId;
    case RELEASE_RV2 :
      return "RV2 " + ThreadData.ThreadId;
    case NATIVE :
      return String.format( "NATIVE%5d ", Index ) + Receiver.FullName();
    case CHOICE :
      return "CHOICE " + Index;
    case QO :
      return "QO";
    }
    return "TismBUG";
  }
}

class TsmScenario
{
  public static int           ScenarioId = 0;
  public int                  Id;
  public TsmScenario          Parent;
  public Vector<TsmScenItem>  Items = new Vector<TsmScenItem>();
  public TsmSnapshot          Snapshot = null;

  public TsmScenario()
  {
    ScenarioId = 0;
    Id = ScenarioId++;
    Parent = null;
  }

  public TsmScenario( TsmScenario oParent )
  {
    Id = ScenarioId++;
    Parent = oParent;
  }

  public int getParentId()
  {
    return ( Parent == null ) ? -1 : Parent.Id;
  }

  public TsmScenario Extend( TsmScenItem oTsmScenItem )
  {
    TsmScenario lTsmScenario = new TsmScenario( this );

    lTsmScenario.Items = new Vector<TsmScenItem>( this.Items );
    lTsmScenario.Items.add( oTsmScenItem );
    lTsmScenario.Snapshot = this.Snapshot;

    return lTsmScenario;
  }

  public void Add( TsmScenItem oTsmScenItem )
  {
    Items.add( oTsmScenItem );
  }

  public void Update( TsmSnapshot oSnapshot )
  {
    oSnapshot.Parent = this.Snapshot;
    this.Snapshot = oSnapshot;
  }

  public boolean HasCycle()
  {
    TsmSnapshot lSnapshot = this.Snapshot;

    while ( null != (lSnapshot = lSnapshot.Parent) )
    {
      if ( lSnapshot.equals( this.Snapshot ) )
      {
        return true;
      }
    }
    return false;
  }

  @Override
  public String toString()
  {
    StringBuilder sb = new StringBuilder();
    int i = 0;
    for ( TsmScenItem tItem : Items )
    {
      sb.append( String.format( "%3d ", i++ ) );
      sb.append( tItem.toString() );
      sb.append( Tsm.NewLine );
    }
    return sb.toString();
  }
}

class TsmSnapshot
{
  public TsmSnapshot  Parent = null;
  public String       SnapClasses;
  public String       SnapCbThread;
  public String       SnapThreads;
  private int         MyHash = 0;

  public TsmSnapshot( String sSnapClasses, String sSnapCbThread, String sSnapThreads )
  {
    SnapClasses  = sSnapClasses;
    SnapCbThread = sSnapCbThread;
    SnapThreads  = sSnapThreads;

    MyHash = toString().hashCode();
  }

  @Override
  public String toString()
  {
    return "{" + SnapClasses + " &" + SnapCbThread + " &" + SnapThreads + "}";
  }

  @Override
  public boolean equals( Object o2 )
  {
    TsmSnapshot ss2 = (TsmSnapshot) o2;
    return (this.SnapClasses.equals(ss2.SnapClasses)
        &&  this.SnapCbThread.equals(ss2.SnapCbThread)
        &&  this.SnapThreads.equals(ss2.SnapThreads) );
  }

  @Override
  public int hashCode()
  {
    return MyHash;
  }
}

class TsmSnapdata
{
  public int          NrRecords;
  public int          NrCalls;
  public TsmScenario  Scenario;

  public TsmSnapdata( int iNrRecords, int iNrCalls, TsmScenario oScenario )
  {
    NrRecords = iNrRecords;
    NrCalls = iNrCalls;
    Scenario = oScenario;
  }
}

class TsmThreadData
{
  public TsmThreadType  Type;
  public int            ThreadId;
  public TsmThread      TsmThread;
  public TsmThreadState State;

  public boolean hasMessage()
  {
    return TsmThread.hasMessage();
  }

  public String toState()
  {
    return( "Thread " + this.ThreadId
          + " Type " + this.Type
          + " State " + this.State
          + " ThreadState " + this.TsmThread.getThread().getState() );
    }    
  }

/*
 * Received client natives have negative values.
 */
class TsmNativeLists
{
  public Vector<Integer>  ClientNatives = new Vector<Integer>();
  public Vector<Integer>  ServerNatives = new Vector<Integer>();
}

public class TsmSupervisor
{
  private static TsmTriggerSem mTriggerSem = new TsmTriggerSem();
  private static TsmActivity   mActivity;
  private static int           mThreadId = 0;
  private static TsmCbThread   mCbThread;
  private static TsmThreadData MyThreadData;
  private static TsmThreadData CbThreadData;
  private static HashMap<Thread,TsmThreadData> mThreadDatas = new HashMap<Thread,TsmThreadData>();
  
  public static void Initialise()
  {
    Tsm.Open();
    mActivity = new TsmActivity();
    Register( null, TsmThreadType.SUPERVISOR );

    mCbThread = new TsmCbThread();
    mCbThread.Prepare();
    mCbThread.Run();
  }

  public static void Terminate()
  {
    mCbThread.Halt();
    mCbThread.Cease();
    Tsm.Close();
  }

  public static TsmIntQueue GetIntQueue()
  {
    return mCbThread.getIntQueue();
  }

  public static void Announce()
  {
    mActivity.Announce();
  }

  public static void Abandon()
  {
    mActivity.Abandon();
  }

  public static int Register( TsmThread oTsmThread )
  {
    return Register( oTsmThread, TsmThreadType.THREAD );
  }

  public static int Register( TsmThread oTsmThread, TsmThreadType eType )
  {
    TsmThreadData lThreadData;
    Thread oThread;
    
    if ( null != oTsmThread )
    {
      oThread = oTsmThread.getThread();
    }
    else
    {
      oThread = Thread.currentThread();
    }

    if ( mThreadDatas.containsKey( oThread ) )
    {
      throw new TsmException( "TismBUG: Doubly registered thread" );
    }

    lThreadData = new TsmThreadData();
    lThreadData.Type      = eType;
    lThreadData.ThreadId  = mThreadId++;
    lThreadData.TsmThread = oTsmThread;
    lThreadData.State     = TsmThreadState.RUNNING;

    switch ( lThreadData.Type )
    {
    case SUPERVISOR:
      MyThreadData = lThreadData;
      break;
    case CB_THREAD:
      CbThreadData = lThreadData;
      break;
    case THREAD:
      mThreadDatas.put( oThread, lThreadData );
      break;
    }
   
    return lThreadData.ThreadId;
  }

  public static void SetState( Thread oThread )
  {
    SetState( oThread, TsmThreadState.RUNNING );
  }

  public static void SetState( TsmThreadState eState )
  {
    SetState( Thread.currentThread(), eState );
  }

  public static void SetState( Thread lThread, TsmThreadState eState )
  {
    TsmThreadData lThreadData;

    if ( lThread == CbThreadData.TsmThread.getThread() )
    {
      lThreadData = CbThreadData;
    }
    else
    {
      lThreadData = mThreadDatas.get( lThread );
      if ( null == lThreadData )
      {
        throw new TsmException( "TismBUG: Unregistered thread" );
      }
    }

    lThreadData.State = eState;
    switch ( lThreadData.Type )
    {
    case SUPERVISOR:
      switch ( lThreadData.State )
      {
      case RUNNING:
      case INTERRUPTED:
        throw new TsmException( "TismBUG: Supervisor should not report" );
        //break;
      case WAITING_MSG:
      case WAITING_RV:
      case WAITING_RV2:
        throw new TsmException( "TismBUG: Supervisor should not wait" );
        //break;
      }
      break;
    case CB_THREAD:
    case THREAD:
      switch ( lThreadData.State )
      {
      case RUNNING:
        break;
      case WAITING_MSG:
        mTriggerSem.Release();
        Pause( lThreadData );
        break;
      case WAITING_RV:
        mTriggerSem.Release();
        break;
      case WAITING_RV2:
        mTriggerSem.Release();
        Pause( lThreadData );
        break;
      case INTERRUPTED:
        lThreadData.State = TsmThreadState.RUNNING;
        break;
      }
      break;
    }
  }

  private static void CheckAllThreadsWaiting()
  {
    while ( ! AllThreadsWaiting() )
    {
      mTriggerSem.WaitOne();
    }

    while ( ! AllThreadsSleeping() )
    {
      try
      {
        Thread.sleep( 1 );
      }
      catch ( InterruptedException e )
      {
      }
    }
  }

  private static boolean AllThreadsWaiting()
  {
    if ( TsmThreadState.RUNNING == CbThreadData.State )
    {
      return false;
    }

    Set<Map.Entry<Thread,TsmThreadData>> entries = mThreadDatas.entrySet(); 
    Iterator<Map.Entry<Thread,TsmThreadData>> iterator = entries.iterator();
    while( iterator.hasNext() )
    { 
      Map.Entry<Thread,TsmThreadData> kvp = (Map.Entry<Thread,TsmThreadData>) iterator.next();
      TsmThreadData lThreadData = kvp.getValue();

      if ( TsmThreadState.RUNNING == lThreadData.State )
      {
        return false;
      }
    }
    return true;
  }

  private static boolean AllThreadsSleeping()
  {
    if ( CbThreadData.TsmThread.getThread().getState() != Thread.State.WAITING )
    {
      return false;
    }

    Set<Map.Entry<Thread,TsmThreadData>> entries = mThreadDatas.entrySet(); 
    Iterator<Map.Entry<Thread,TsmThreadData>> iterator = entries.iterator();
    while( iterator.hasNext() )
    { 
      Map.Entry<Thread,TsmThreadData> kvp = (Map.Entry<Thread,TsmThreadData>) iterator.next();
      TsmThreadData lThreadData = kvp.getValue();

      if ( lThreadData.TsmThread.getThread().getState() != Thread.State.WAITING )
      {
        return false;
      }
    }
    return true;
  }

  private synchronized static String GetThreadStates( String sHeader )
  {
    StringBuilder sb = new StringBuilder();
    TsmThreadData lThreadData;

    sb.append( sHeader );
    sb.append( Tsm.NewLine );
    sb.append( MyThreadData.toState() );
    sb.append( Tsm.NewLine );
    sb.append( CbThreadData.toState() );
    sb.append( Tsm.NewLine );

    Set<Map.Entry<Thread,TsmThreadData>> entries = mThreadDatas.entrySet(); 
    Iterator<Map.Entry<Thread,TsmThreadData>> iterator = entries.iterator();
    while( iterator.hasNext() )
    { 
      Map.Entry<Thread,TsmThreadData> kvp = (Map.Entry<Thread,TsmThreadData>) iterator.next();
      lThreadData = kvp.getValue();

      sb.append( lThreadData.toState() );
      sb.append( Tsm.NewLine );
    }
    return sb.toString();
  }
  
  private static Hashtable<TsmClass,TsmNativeLists> Natives = new Hashtable<TsmClass,TsmNativeLists>();

  private static void NativesReset()
  {
    Natives.clear();
  }

  public static void NativeEntry( TsmClass oOwner, int [] aNatives )
  {
    synchronized( Natives )
    {
      TsmNativeLists lLists;
  
      lLists = Natives.get( oOwner );
      if ( null == lLists )
      {
        lLists = new TsmNativeLists();
        Natives.put( oOwner, lLists );
      }
  
      for ( int n : aNatives )
      {
        if ( n < 0 )
        {
          lLists.ClientNatives.add( (Integer) n );
        }
        else
        {
          lLists.ServerNatives.add( (Integer) n );
        }
      }
    }
  }

  public static void NativeExit( TsmClass oOwner, int [] aNatives )
  {
    synchronized( Natives )
    {
      TsmNativeLists lLists;
  
      lLists = Natives.get( oOwner );
      if ( null == lLists )
      {
        throw new TsmException( "TismBUG: NativeExit " + oOwner );
      }
  
      for ( int n : aNatives )
      {
        if ( n < 0 )
        {
          lLists.ClientNatives.remove( (Integer) n );
        }
        else
        {
          lLists.ServerNatives.remove( (Integer) n );
        }
      }
    }
  }

  public static HashMap<TsmSnapshot,TsmSnapdata>  Snapshots = new HashMap<TsmSnapshot,TsmSnapdata>();
  public static Vector<TsmClass>        Classes = new Vector<TsmClass>();
  public static Vector<TsmScenario>     PastScenarios = new Vector<TsmScenario>();
  public static Stack<TsmScenario>    FutureScenarios = new Stack<TsmScenario>();
  public static TsmScenario           CurrentScenario;
  public static int                   CurrentScenItemIndex = 0;

  private static int          NumberOfDeadlocks = 0;
  private static int          NumberOfLivelocks = 0;
  private static int          NumberOfCycles = 0;
  private static long         NumberOfLoops = 0;
  private static long         StartTime;
  private static long         StopTime;

  public static String GetCurrentClasses()
  {
    StringBuilder sb = new StringBuilder();
    int c = 0;
    for( TsmClass tClass : Classes )
    {
      sb.append( String.format( "%3d ", ++c ) );
      sb.append( tClass.TsmSnap() );
      sb.append( Tsm.NewLine );
    }
    return sb.toString();
  }

  public static String GetCurrentScenario()
  {
    return PastScenarios.lastElement().toString();
  }

  private static void Reset()
  {
    mCbThread.TsmReset1();

    Set<Map.Entry<Thread,TsmThreadData>> entries = mThreadDatas.entrySet(); 
    Iterator<Map.Entry<Thread,TsmThreadData>> iterator = entries.iterator();
    while( iterator.hasNext() )
    { 
      Map.Entry<Thread,TsmThreadData> kvp = (Map.Entry<Thread,TsmThreadData>) iterator.next();
      TsmThreadData lThreadData = kvp.getValue();
      lThreadData.TsmThread.TsmReset1();
    }

    if ( TsmThreadState.WAITING_RV == CbThreadData.State )
    {
      Tsm.Lgl( 7, "Interrupt Cb" );
      CbThreadData.State = TsmThreadState.RUNNING;
      mCbThread.Interrupt();
    }

    entries = mThreadDatas.entrySet(); 
    iterator = entries.iterator();
    while( iterator.hasNext() )
    { 
      Map.Entry<Thread,TsmThreadData> kvp = (Map.Entry<Thread,TsmThreadData>) iterator.next();
      TsmThreadData lThreadData = kvp.getValue();

      if ( TsmThreadState.WAITING_RV == lThreadData.State )
      {
        Tsm.Lgl( 7, "Interrupt T_" + lThreadData.TsmThread.mSequence );
        lThreadData.State = TsmThreadState.RUNNING;
        lThreadData.TsmThread.Interrupt();
      }
    }
    CheckAllThreadsWaiting();

    mCbThread.TsmReset2();

    entries = mThreadDatas.entrySet(); 
    iterator = entries.iterator();
    while( iterator.hasNext() )
    { 
      Map.Entry<Thread,TsmThreadData> kvp = (Map.Entry<Thread,TsmThreadData>) iterator.next();
      TsmThreadData lThreadData = kvp.getValue();
      lThreadData.TsmThread.TsmReset2();
    }

    for ( TsmClass tClass : Classes )
    {
      tClass.TsmReset1();
    }
    NativesReset();
  }

  private static void Free()
  {
    Set<Map.Entry<Thread,TsmThreadData>> entries = mThreadDatas.entrySet(); 
    Iterator<Map.Entry<Thread,TsmThreadData>> iterator = entries.iterator();
    while( iterator.hasNext() )
    { 
      Map.Entry<Thread,TsmThreadData> kvp = (Map.Entry<Thread,TsmThreadData>) iterator.next();
      TsmThreadData lThreadData = kvp.getValue();
      lThreadData.TsmThread.Halt();
    }
  }
  
  private static void Pause( TsmThreadData oThreadData )
  {
    oThreadData.TsmThread.TriggerWait();
  }

  private static void Continue( TsmThreadData oThreadData )
  {
    oThreadData.State = TsmThreadState.RUNNING;
    oThreadData.TsmThread.TriggerNotify();
  }

  private static void Replay()
  {
    for ( TsmClass tClass : Classes )
    {
      tClass.TsmReset2();
    }

    CurrentScenItemIndex = 0;
    TsmRecorder.Reset();

    for ( ; ; )
    {
      CheckAllThreadsWaiting();

      if ( CurrentScenItemIndex == CurrentScenario.Items.size() )
      { /*
         * Ready replaying, possibly due to choice items, or
         * empty scenario.
         */
        TsmRecorder.Enable();
        break;
        /****/
      }
      if ( CurrentScenItemIndex == CurrentScenario.Items.size()-1 )
      { /*
         * The last item is 'new'.
         */
        TsmRecorder.Enable();
      }
      TsmScenItem lScenItem = CurrentScenario.Items.get( CurrentScenItemIndex++ );
      ReplayScenItem( lScenItem );
    }
  }

  private static void ReplayNative( TsmScenItem oItem )
  {
    Tsm.Lgl( 6, "Natv: " + oItem.Index + " @ " + oItem.Receiver.FullName() );
    if ( oItem.Index < 0 )
    { /*
       * Client native.
       * Executed by CbThread.
       */
      TsmTrigger lTrigger = new TsmTrigger( -oItem.Index );
      TsmEnvelope lEnvelope = new TsmEnvelope( lTrigger, oItem.Receiver );
      mCbThread.NativeCall( lEnvelope );
      Continue( CbThreadData );
    }
    else
    { /*
       * Server native.
       * Executed by Supervisor.
       */
      oItem.Receiver.ExecuteNative( oItem.Index );
    }
  }

  /*
   * The state changes :
   *     Item.ThreadData.State = TsmThreadState.RUNNING;
   * are needed to force the Supervisor to wait for the threads.
   */
  private static void ReplayScenItem( TsmScenItem Item )
  {
    switch ( Item.Kind )
    {
    case RELEASE_CB_MSG :
      //Tsm.Assert( Item.ThreadData.hasMessage() );
      Continue( Item.ThreadData );
      break;
    case RELEASE_CB_RV2 :
      //Tsm.Assert( Item.ThreadData.State == TsmThreadState.WAITING_RV2 );
      Continue( Item.ThreadData );
      break;
    case RELEASE_MSG :
      //Tsm.Assert( Item.ThreadData.hasMessage() );
      Continue( Item.ThreadData );
      break;
    case RELEASE_RV2 :
      //Tsm.Assert( ( Item.ThreadData.State & TsmThreadState.WAITING_RV ) != 0 );
      Continue( Item.ThreadData );
      break;
    case NATIVE :
      ReplayNative( Item );
      break;
    case CHOICE :
      TsmRecorder.Enable();
      TsmRecorder.Show( "**** BUG " );
      throw new TsmException( "TismBUG: Handled by 'ChoiceMax'" );
      //break;
    }
  }

  private static TsmScenItem GetThreadDataRV2()
  {
    TsmScenItem lScenItem;

    if ( TsmThreadState.WAITING_RV2 == CbThreadData.State )
    {
      lScenItem = new TsmScenItem();
      lScenItem.Kind = TsmScenItemKind.RELEASE_CB_RV2;
      lScenItem.ThreadData = CbThreadData;
      return lScenItem;
    }

    Set<Map.Entry<Thread,TsmThreadData>> entries = mThreadDatas.entrySet(); 
    Iterator<Map.Entry<Thread,TsmThreadData>> iterator = entries.iterator();
    while( iterator.hasNext() )
    { 
      Map.Entry<Thread,TsmThreadData> kvp = (Map.Entry<Thread,TsmThreadData>) iterator.next();
      TsmThreadData lThreadData = kvp.getValue();

      if ( TsmThreadState.WAITING_RV2 == lThreadData.State )
      {
        lScenItem = new TsmScenItem();
        lScenItem.Kind = TsmScenItemKind.RELEASE_RV2;
        lScenItem.ThreadData = lThreadData;
        return lScenItem;
      }
    }
    
    return null;
  }

  private static void ReplayExtend( TsmScenario oScenario )
  {
    TsmScenItem lScenItem;

    while ( null != (lScenItem = GetThreadDataRV2()) )
    {
      oScenario.Add( lScenItem );
      CurrentScenItemIndex++;
      ReplayScenItem( lScenItem );
      CheckAllThreadsWaiting();
    }
  }

  private static boolean  mbQueueOverflow = false;
  private static int      mMaximumQueueSize = 7;
  /*
   * Called by Queues.
   */
  public synchronized static boolean QueueOverflow( int QueueCount )
  {
    if ( QueueCount >= mMaximumQueueSize )
    {
      TsmScenItem lScenItem = new TsmScenItem();
      lScenItem.Kind = TsmScenItemKind.QO;
      CurrentScenario.Add( lScenItem );
      CurrentScenItemIndex++;
      mbQueueOverflow = true;
    }
    return mbQueueOverflow;
  }

  private static boolean  mbVerify = true;
  private static Random   mRandom = new Random( 1 );
  private static int      mRandomSeed = 1;
  private static long     mMaxNrSimLoops = 1000;
  /*
   * Called by Fsms.
   */
  public synchronized static boolean ChoiceOK( int MaxNrChoices, int ChoiceIndex, String sGuard )
  {
    if ( mbVerify )
    {
      return ChoiceOKVerify( MaxNrChoices, ChoiceIndex, sGuard );
    }

    return ChoiceOKSimulate( MaxNrChoices, ChoiceIndex );
  }

  private static boolean ChoiceOKVerify( int MaxNrChoices, int ChoiceIndex, String sGuard )
  {
    TsmScenItem lScenItem;

    if ( CurrentScenItemIndex < CurrentScenario.Items.size() )
    { /*
       * Still replaying.
       */
      lScenItem = CurrentScenario.Items.get( CurrentScenItemIndex );
      //Tsm.Assert( lScenItem.Kind == TsmScenItemKind.CHOICE );
      if ( lScenItem.Index != ChoiceIndex )
      {
        return false;
        /***********/
      }

      CurrentScenItemIndex++;
      if ( CurrentScenItemIndex == CurrentScenario.Items.size() )
      { /*
         * 'Execution' of the last item is happening now.
         */
        TsmRecorder.Enable();
      }
    }
    else
    { /*
       * The scenario has been replayed fully, and
       * now encounters a Choice node.
       */
      //Tsm.Assert( ChoiceIndex == 1 );
      TsmScenario lNewScenario;
      for ( int i=2; i<=MaxNrChoices; i++ )
      {
        lScenItem = new TsmScenItem();
        lScenItem.Kind = TsmScenItemKind.CHOICE;
        lScenItem.Index = i;
        lNewScenario = CurrentScenario.Extend( lScenItem );
        FutureScenarios.push( lNewScenario );
      }

      lScenItem = new TsmScenItem();
      lScenItem.Kind = TsmScenItemKind.CHOICE;
      lScenItem.Index = 1;
      CurrentScenario.Add( lScenItem );
      CurrentScenItemIndex++;
    }
    Tsm.Seq( "H," + sGuard );
    return true;
  }

  public static void Verify()
  {
    Vector<TsmScenItem> lScenItems = new Vector<TsmScenItem>();
    boolean             bProxyProtocol = false;
    TsmSnapshot         lSnapshot;
    TsmSnapdata         lSnapdata;

    mbVerify = true;
    StartTime = System.currentTimeMillis();

    Thread.currentThread().setPriority( Thread.MIN_PRIORITY );

    if ( Classes.size() == 2 )
    {
      if ( ( ( Classes.get(0) instanceof TsmProxy ) && ( Classes.get(1) instanceof TsmProxyUser ) )
        || ( ( Classes.get(1) instanceof TsmProxy ) && ( Classes.get(0) instanceof TsmProxyUser ) ) )
      {
        bProxyProtocol = true;
      }
    }
    CurrentScenario = new TsmScenario();
    FutureScenarios.push( CurrentScenario );

    boolean mbCurrent = false;
    while ( ! FutureScenarios.empty() )
    {
      CurrentScenario = FutureScenarios.pop();
      PastScenarios.add( CurrentScenario );

      if ( mbCurrent )
      {
        mbCurrent = false;

        Tsm.Seq( "S," + CurrentScenario.Id + "," + CurrentScenario.getParentId() );
        Tsm.Lgl( 2, "Scenario [ " + CurrentScenario.Id + "  :  " + CurrentScenario.getParentId() + " ] -------------------------------------------------------" );
        Tsm.Lgl( 3, CurrentScenario.toString() );
        if ( CurrentScenario.Items.size() > 0 )
        {
          CheckAllThreadsWaiting();
          TsmScenItem lScenItem = CurrentScenario.Items.get( CurrentScenItemIndex++ );
          ReplayScenItem( lScenItem );
          //Tsm.Assert( CurrentScenItemIndex == CurrentScenario.Items.size() );
        }
        CheckAllThreadsWaiting();
      }
      else
      {
        Tsm.Seq( "N" );
        Reset();
        Tsm.Seq( "S," + CurrentScenario.Id + "," + CurrentScenario.getParentId() );
        Tsm.Lgl( 2, "Scenario [ " + CurrentScenario.Id + "  :  " + CurrentScenario.getParentId() + " ] -------------------------------------------------------" );
        Tsm.Lgl( 3, CurrentScenario.toString() );
        Replay();
      }
      ReplayExtend( CurrentScenario );
      /*
       * Now, no threads have state RV2 !
       */

      lSnapshot = new TsmSnapshot(
        SnappedClasses(),
        CbThreadData.TsmThread.Snap(),
        SnappedThreads() );
      CurrentScenario.Update( lSnapshot );

      if ( mbQueueOverflow )
      {
        mbQueueOverflow = false;
        TsmRecorder.QueueOverflow();
        continue;
        /*******/
      }

      Tsm.Seq( "T," + lSnapshot.SnapClasses );
      if ( null != ( lSnapdata=Snapshots.get( lSnapshot ) ) )
      {
        Tsm.Lgl( 5, "Equal: " + lSnapdata.Scenario.Id );
        Tsm.Seq( "Q," + lSnapdata.Scenario.Id );
        if ( CurrentScenario.HasCycle() )
        {
          NumberOfCycles++;
          Tsm.Lgl( 4, "Cycle: -------------------------------------------------------" );
          Tsm.Seq( "Y" );
          if ( lSnapdata.NrCalls == TsmRecorder.getNumberOfExtCalls() )
          { /*
             * The number of external calls retrieved from the previous
             * position in lSnapdata is equal to the current number of
             * external calls in the Recorder.
             */
            NumberOfLivelocks++;
            TsmRecorder.Show( "**** LIVELOCK ", lSnapdata.NrRecords );
          }
        }
      }
      else
      {
        Tsm.Lgl( 5, "Rply: -------------------------------------------------------" );
        Tsm.Lgl( 5, "Snap:" + lSnapshot.toString() );
        lSnapdata = new TsmSnapdata( TsmRecorder.getNumberOfRecords(), TsmRecorder.getNumberOfExtCalls(), CurrentScenario );
        Snapshots.put( lSnapshot, lSnapdata );

        boolean bClient = (CbThreadData.State == TsmThreadState.WAITING_MSG);
        lScenItems.clear();
        AddScenarioCbThread( lScenItems );
        if ( bProxyProtocol )
        { /*
           * In case of a proxy protocol, native triggers are only
           * allowed in the absense of pending callbacks.
           */
          if ( lScenItems.size() == 0 )
          {
            AddScenarioNatives( lScenItems, bClient );
            AddScenarioThreads( lScenItems );
          }
        }
        else
        {
          AddScenarioNatives( lScenItems, bClient );
          AddScenarioThreads( lScenItems );
        }

        if ( lScenItems.size() == 0 )
        {
          NumberOfDeadlocks++;
          TsmRecorder.Show( "**** DEADLOCK " );
        }
        else
        {
          AddToCurrentScenario( lScenItems );
          mbCurrent = true;
        }
      }
    }

    /*
     * Cleanup pending threads.
     */
    Reset();
    Free();
    StopTime = System.currentTimeMillis();
  }

  public static void ConfigureSimulation( int iSeed, long iNrLoops )
  {
    mRandomSeed = iSeed;
    mMaxNrSimLoops = iNrLoops;
  }

  private static boolean ChoiceOKSimulate( int MaxNrChoices, int ChoiceIndex )
  {
    return (mRandom.nextInt( MaxNrChoices + 1 - ChoiceIndex ) == 0);
  }

  public static void Simulate()
  {
    Vector<TsmScenItem> lScenItems = new Vector<TsmScenItem>();
    boolean              bProxyProtocol = false;

    mbVerify = false;
    mRandom = new Random( mRandomSeed );
    StartTime = System.currentTimeMillis();

    Thread.currentThread().setPriority( Thread.MIN_PRIORITY );

    if ( Classes.size() == 2 )
    {
      if ( ( ( Classes.get(0) instanceof TsmProxy ) && ( Classes.get(1) instanceof TsmProxyUser ) )
        || ( ( Classes.get(1) instanceof TsmProxy ) && ( Classes.get(0) instanceof TsmProxyUser ) ) )
      {
        bProxyProtocol = true;
      }
    }

    boolean bCurrent = false;
    NumberOfLoops = 0;
    do
    {
      if ( ++NumberOfLoops == 0 )
      {
        NumberOfLoops = 1;
      }

      Tsm.Lgl( 3, "Loop: " + NumberOfLoops );
      if ( bCurrent )
      {
        bCurrent = false;

        CheckAllThreadsWaiting();
        int r = mRandom.nextInt( lScenItems.size() );
        CurrentScenario.Extend( lScenItems.get(r) );
        ReplayScenItem( lScenItems.get(r) );
        CheckAllThreadsWaiting();
      }
      else
      {
        Reset();
        CurrentScenario = new TsmScenario();
        Replay();
      }

      TsmScenItem lScenItem;
      while ( null != ( lScenItem = GetThreadDataRV2() ) )
      {
        ReplayScenItem( lScenItem );
        CheckAllThreadsWaiting();
      }

      boolean bClient = ( CbThreadData.State == TsmThreadState.WAITING_MSG );
      lScenItems.clear();
      AddScenarioCbThread( lScenItems );
      if ( bProxyProtocol )
      { /*
         * In case of a proxy protocol, native triggers are only
         * allowed in the absense of pending callbacks.
         */
        if ( lScenItems.size() == 0 )
        {
          AddScenarioNatives( lScenItems, bClient );
          AddScenarioThreads( lScenItems );
        }
      }
      else
      {
        AddScenarioNatives( lScenItems, bClient );
        AddScenarioThreads( lScenItems );
      }

      if ( lScenItems.size() == 0 )
      {
        NumberOfDeadlocks++;
        TsmRecorder.Show( "**** DEADLOCK " );
      }
      else
      {
        bCurrent = true;
      }
    } while ( NumberOfLoops != mMaxNrSimLoops );

    Reset();
    Free();
    StopTime = System.currentTimeMillis();
  }

  public static String SnappedClasses()
  {
    StringBuilder sb = new StringBuilder();
    int c = 0;
    for( TsmClass tClass : Classes )
    {
      sb.append( "" + ++c  );
      sb.append( tClass.TsmSnap() );
    }
    return sb.toString();
  }

  public static String SnappedThreads()
  {
    StringBuilder sb = new StringBuilder();
    int t = 0;
    Set<Map.Entry<Thread,TsmThreadData>> entries = mThreadDatas.entrySet(); 
    Iterator<Map.Entry<Thread,TsmThreadData>> iterator = entries.iterator();
    while( iterator.hasNext() )
    { 
      Map.Entry<Thread,TsmThreadData> kvp = (Map.Entry<Thread,TsmThreadData>) iterator.next();
      TsmThreadData lThreadData = kvp.getValue();

      sb.append( " T_" + ++t + ": " );
      sb.append( lThreadData.TsmThread.Snap() );
    }
    return sb.toString();
  }

  private static void AddToCurrentScenario( Vector<TsmScenItem> oScenItems )
  {
    for( TsmScenItem si : oScenItems )
    {
      TsmScenario lNewScenario = CurrentScenario.Extend( si );
      FutureScenarios.push( lNewScenario );
    }
  }

  private static void AddScenarioNatives( Vector<TsmScenItem> oScenItems, boolean bClient )
  {
    synchronized( Natives )
    {
      Set<Map.Entry<TsmClass,TsmNativeLists>> entries = Natives.entrySet(); 
      Iterator<Map.Entry<TsmClass,TsmNativeLists>> iterator = entries.iterator();
      while( iterator.hasNext() )
      { 
        Map.Entry<TsmClass,TsmNativeLists> kvp = (Map.Entry<TsmClass,TsmNativeLists>) iterator.next();
        TsmNativeLists lLists = kvp.getValue();
        TsmClass Receiver = kvp.getKey();
  
        for ( int index : lLists.ServerNatives )
        {
          TsmScenItem lScenItem = new TsmScenItem();
          lScenItem.Kind = TsmScenItemKind.NATIVE;
          lScenItem.Index = index;
          lScenItem.Receiver = Receiver;
          oScenItems.add( lScenItem );
        }
  
        if ( bClient )
        {
          for ( int index : lLists.ClientNatives )
          {
            TsmScenItem lScenItem = new TsmScenItem();
            lScenItem.Kind = TsmScenItemKind.NATIVE;
            lScenItem.Index = index;
            lScenItem.Receiver = Receiver;
            oScenItems.add( lScenItem );
          }
        }
      }
    }
  }

  private static void AddScenarioCbThread( Vector<TsmScenItem> oScenItems )
  {
    if ( CbThreadData.TsmThread.getIntQueue().size() == 0 )
    {
      return;
    }

    if ( CbThreadData.State == TsmThreadState.WAITING_RV )
    {
      return;
    }
    //Tsm.Assert( CbThreadData.State == TsmThreadState.WAITING_MSG );

    TsmScenItem lScenItem = new TsmScenItem();
    lScenItem.Kind = TsmScenItemKind.RELEASE_CB_MSG;
    lScenItem.ThreadData = CbThreadData;
    oScenItems.add( lScenItem );
  }

  private static void AddScenarioThreads( Vector<TsmScenItem> oScenItems )
  {
    Set<Map.Entry<Thread,TsmThreadData>> entries = mThreadDatas.entrySet(); 
    Iterator<Map.Entry<Thread,TsmThreadData>> iterator = entries.iterator();
    while( iterator.hasNext() )
    { 
      Map.Entry<Thread,TsmThreadData> kvp = (Map.Entry<Thread,TsmThreadData>) iterator.next();
      TsmThreadData lThreadData = kvp.getValue();

      if ( TsmThreadState.WAITING_MSG == lThreadData.State )
      {
        if ( lThreadData.TsmThread.hasMessage() )
        {
          TsmScenItem lScenItem = new TsmScenItem();
          lScenItem.Kind = TsmScenItemKind.RELEASE_MSG;
          lScenItem.ThreadData = lThreadData;
          oScenItems.add( lScenItem );
        }
      }
    }
  }

  private static String GetDuration( long duration )
  {
    duration /= 1000;
    long s = duration % 60;
    duration /= 60;
    long m = duration % 60;
    duration /= 60;
    return String.format( "%02d:%02d:%02d", duration, m, s );
  }

  public static void Summary()
  {
    long duration = StopTime - StartTime;
    Tsm.Log( "" );
    Tsm.Log( "SUMMARY" );
    Tsm.Log( "-------" );
    if ( mbVerify )
    {
      Tsm.Log( "Number of scenarios: " + PastScenarios.size() );
      Tsm.Log( "Number of cycles   : " + NumberOfCycles );
    }
    else
    {
      Tsm.Log( "Number of loops    : " + NumberOfLoops );
    }
    Tsm.Log( "Number of deadlocks: " + NumberOfDeadlocks );
    if ( mbVerify )
    {
      Tsm.Log( "Number of livelocks: " + NumberOfLivelocks );
    }
    Tsm.Log( "Number of unhandled functions  : " + TsmRecorder.NumberOfUnhandledFunctions );
    Tsm.Log( "Number of unvisited states     : " + TsmRecorder.NumberOfUnvisitedStates );
    Tsm.Log( "Number of unvisited transitions: " + TsmRecorder.NumberOfUnvisitedTransitions );
    Tsm.Log( "Number of rendez-vous errors   : " + TsmRecorder.NumberOfSyncReturnErrors );
    Tsm.Log( "Number of queue overflows      : " + TsmRecorder.NumberOfQueueOverflows );
    Tsm.Log( "Duration: " + GetDuration( duration ) );
    Tsm.Log( "-------" );
    Tsm.Log( "Ext  Int  Thread's maximum queue sizes" );
    Tsm.Log( mCbThread.MaxQueueSizes() );
    Set<Map.Entry<Thread,TsmThreadData>> entries = mThreadDatas.entrySet(); 
    Iterator<Map.Entry<Thread,TsmThreadData>> iterator = entries.iterator();
    while( iterator.hasNext() )
    { 
      Map.Entry<Thread,TsmThreadData> kvp = (Map.Entry<Thread,TsmThreadData>) iterator.next();
      TsmThreadData lThreadData = kvp.getValue();

      Tsm.Log( lThreadData.TsmThread.MaxQueueSizes() );
    }
    Tsm.Log( "-------" );
  }
}
