/*
 * TismTool © Tis Veugen, 2010-2012
 * Simulation, Verification
 *
 * TsmThread.java
 * version 5
 */
package TsmRTE.TsmSupervisor;

import TsmRTE.TsmGen.Tsm;
import TsmRTE.TsmGen.TsmEnvelope;
import TsmRTE.TsmGen.TsmRvState;

public class TsmThread
{
  public TsmThreadType    mType;
  public int              mSequence;
  protected TsmWorker     mThread;
  protected TsmExtQueue   mExternalQueue = new TsmExtQueue();
  protected TsmIntQueue   mInternalQueue;
  protected TsmMutex      mExclusiveMutex = new TsmMutex();
  protected TsmRzVousSem  mRendezVousSem = new TsmRzVousSem();
  protected TsmTriggerSem mTriggerSem = new TsmTriggerSem();
  protected Thread        mRendezVousThread;

  public TsmThread()
  {
    this( TsmThreadType.THREAD );
  }

  public TsmThread( TsmThreadType Type )
  {
    mType = Type;
    mInternalQueue = new TsmIntQueue();
  }

  public Thread getThread()
  {
     return mThread;
  }

  public TsmExtQueue getExtQueue()
  {
    return mExternalQueue;
  }

  public TsmIntQueue getIntQueue()
  {
    return mInternalQueue;
  }

  public int getMaxExtQueueSize()
  {
    return mExternalQueue.getMaxQueueSize();
  }

  public int getMaxIntQueueSize()
  {
    return mInternalQueue.getMaxQueueSize();
  }

  public synchronized boolean hasMessage()
  {
    return ( mExternalQueue.size() + mInternalQueue.size() > 0 );
  }

  public String MaxQueueSizes()
  {
    return String.format( "%3d  %3d  %s"
      , this.getMaxExtQueueSize()
      , this.getMaxIntQueueSize()
      , this.mThread.getName() );
  }

  public void Prepare( String sName )
  {
    mThread = new TsmWorker( this );
    mSequence = TsmSupervisor.Register( this, mType );
    mExternalQueue.setSeqId( mSequence );
    mInternalQueue.setSeqId( mSequence );
    mThread.setName( "Tism" + mSequence + "_" + sName );

    TsmSupervisor.Announce();
  }

  public void Run()
  {
    mThread.start();
  }

  public void Halt()
  {
    mThread.MyHalt();
    TriggerNotify();
    try
    {
      mThread.join();
    }
    catch ( InterruptedException e )
    {
      e.printStackTrace();
    }
  }
  
  public void Interrupt()
  {
    mThread.interrupt();
  }

  public void Cease()
  {
  }

  public void ExclusiveLock()
  {
    mExclusiveMutex.WaitOne();
  }

  public void ExclusiveUnlock()
  {
    mExclusiveMutex.ReleaseMutex();
  }

  public void ExclusiveReset()
  {
    mExclusiveMutex = new TsmMutex();
  }

  public void RendezVousWait()
  {
    mRendezVousThread = Thread.currentThread();
    mRendezVousSem.WaitOne();
  }

  public void RendezVousNotify()
  {
    Tsm.Seq( "R," + this.mExternalQueue.getSeqId() );
    TsmSupervisor.SetState( mRendezVousThread );
    mRendezVousSem.Release();
  }

  public void TriggerWait()
  {
    mTriggerSem.WaitOne();
  }

  public void TriggerNotify()
  {
    mTriggerSem.Release();
  }

  public void ASyncCall( TsmEnvelope oEnvelope )
  {
    mExternalQueue.Put( oEnvelope );
  }

  public void SyncCall( TsmEnvelope oEnvelope )
  {
    ExclusiveLock();

    oEnvelope.getTrigger().RvState = TsmRvState.WAIT;
    mExternalQueue.Put( oEnvelope );

    TsmSupervisor.SetState( TsmThreadState.WAITING_RV );
    RendezVousWait();
    TsmSupervisor.SetState( TsmThreadState.WAITING_RV2 );

    ExclusiveUnlock();
  }

  protected TsmRvState Handle( TsmEnvelope oEnvelope )
  throws InterruptedException
  {
    oEnvelope.getTarget().ExecuteTrigger( oEnvelope.getTrigger() );
    if ( oEnvelope.getTrigger().RvState == TsmRvState.READY )
    {
      RendezVousNotify();
      oEnvelope.getTrigger().RvState = TsmRvState.IDLE;
    }
    return oEnvelope.getTrigger().RvState;
  }

  public void CbHandle( TsmEnvelope oEnvelope )
  throws InterruptedException
  {
    // To be overridden!
  }

  public void Prepare()
  {
  }

  public void TsmReset1()
  {
    mExternalQueue.clear();
    mInternalQueue.clear();
  }

  public void TsmReset2()
  {
    mRendezVousSem.Reset();
    ExclusiveReset();
  }

  public String Snap()
  {
    StringBuilder sb = new StringBuilder();

    if ( mExternalQueue.size() > 0 )
    {
      sb.append( " Q:" );
    }
    sb.append( mExternalQueue.Snap() );

    if ( mInternalQueue.size() > 0 )
    {
      sb.append( " S:" );
    }
    sb.append( mInternalQueue.Snap() );

    return sb.toString();
  }
}
